import time


class Stopwatch(object):
    """A stopwatch utility for timing execution that can be used as a regular
    object or as a context manager.
    NOTE: This should not be used an accurate benchmark of Python code, but a
    way to check how much time has elapsed between actions. And this does not
    account for changes or blips in the system clock.

    Instance attributes:

    start_time -- timestamp when the timer started
    stop_time -- timestamp when the timer stopped
    As a regular object:
    >>> stopwatch = Stopwatch()
    >>> stopwatch.start()
    >>> time.sleep(1)
    >>> 1 <= stopwatch.time_elapsed <= 2
    True
    >>> time.sleep(1)
    >>> stopwatch.stop()
    >>> 2 <= stopwatch.total_run_time
    True
    As a context manager:
    >>> with Stopwatch() as stopwatch:
    ...     time.sleep(1)
    ...     print repr(1 <= stopwatch.time_elapsed <= 2)
    ...     time.sleep(1)
    True
    >>> 2 <= stopwatch.total_run_time
    True
    """

    def __init__(self):
        """Initialize a new `Stopwatch`, but do not start timing."""
        self.start_time = None
        self.stop_time = None

    def start(self):
        """Start timing."""
        self.start_time = time.time()

    def stop(self):
        """Stop timing."""
        self.stop_time = time.time()

    @property
    def time_elapsed(self):
        """Return the number of seconds that have elapsed since this
        `Stopwatch` started timing.
        This is used for checking how much time has elapsed while the timer is
        still running.
        """
        assert not self.stop_time, \
            "Can't check `time_elapsed` on an ended `Stopwatch`."
        return time.time() - self.start_time

    @property
    def total_run_time(self):
        """Return the number of seconds that elapsed from when this `Stopwatch`
        started to when it ended.
        """
        return self.stop_time - self.start_time

    def __enter__(self):
        """Start timing and return this `Stopwatch` instance."""
        self.start()
        return self

    def __exit__(self, type, value, traceback):
        """Stop timing.
        If there was an exception inside the `with` block, re-raise it.
        >>> with Stopwatch() as stopwatch:
        ...     raise Exception
        Traceback (most recent call last):
            ...
        Exception
        """
        self.stop()
        if type:
            raise type




from . import io_with_cpp as io
from . import gpu
import matplotlib.pylab as pylab
from subprocess import Popen, PIPE
import sys
from os import path


def get_barnes_hut_executable(exe_dir=None):
    """
    Finds the directory that the Barnes_Hut.exe file would be in if this was installed as a package (or uses the user
    defined folder)

    :param exe_dir: the user defined folder of where the Barnes_Hut.exe file is
    :type exe_dir: string
    :return: the folder where the Barnes_Hut.exe file is
    :rtype: string
    """
    if exe_dir is None:
        exe_file = io.find_exe_file()
        if exe_file is None:
            print('Cannot find Barnes_Hut.exe. Please provide a path to it yourself by setting the exe_dir parameter.')
            return
    else:
        exe_file = path.join(exe_dir, 'Barnes_Hut.exe')

    return exe_file


def run_executable(exe_file, files_dir, verbose):
    """
    Runs the Barnes_Hut.exe executable

    :param exe_file: the full path of the Barnes_Hut executable
    :type exe_file: string
    :param files_dir: the folder where the data.dat file is that the executable is going to read
    :type files_dir: string
    :param verbose: level of verbosity
    :type verbose: int
    """
    with Popen(['Barnes_Hut.exe', ], executable=exe_file, cwd=files_dir, stdout=PIPE, bufsize=1,
               universal_newlines=True) \
            as t_sne_bhcuda_p:
        for line in iter(t_sne_bhcuda_p.stdout):
            print(line, end='')
            sys.stdout.flush()
        t_sne_bhcuda_p.wait()
    assert not t_sne_bhcuda_p.returncode, ('ERROR: Call to Barnes_Hut exited '
                                           'with a non-zero return code exit status, please ' +
                                           ('enable verbose mode and ' if not verbose else '') +
                                           'refer to the t_sne output for further details')


def t_sne(samples, files_dir=None, exe_dir=None, num_dims=2, perplexity=100, theta=0.4, eta=200, exageration=12.0,
          iterations=1000, random_seed=-1, verbose=2):
    """
    Runs the t-SNE algorithm on the samples matrix

    :param samples: the samples x features matrix to run t-SNE on
    :type samples: float32[:,:]
    :param files_dir: where the data.dat file is and where the t-SNE will save its result
    :type files_dir: string
    :param exe_dir: where the Barnes_Hut.exe is (if None then the algorithm will try and find it itself)
    :type exe_dir: string
    :param num_dims: how many dimensions should the t-sne embedding have (2 or 3)
    :type num_dims: int
    :param perplexity: the number of closest spikes considered are 3 * perplexity
    :type perplexity: int
    :param theta: the angle on the t-sne embedding inside of which all spikes are considered a single point (0 to 1)
    :type theta: float
    :param eta: the learning rate of the algorithm
    :type eta: float
    :param exageration: the initial blow out (push out) of the points on the embedding
    :type exageration: float
    :param iterations: the number of iterations of the t-sne algorithm
    :type iterations: int
    :param random_seed: the random seed to create the initial random positions of the points on the embedding
    :type random_seed: float
    :param verbose: levels of verbosity. 0 is no output. If 3 then the algorithm will also save all intermediate t-sne
    embeddings. Useful for making videos of the t-sne process

    :type verbose: int
    :return: the t-NSE result
    :rtype: float32[:,num_of_dims]
    """
    data = pylab.demean(samples, axis=0)
    data /= data.max()
    closest_indices_in_hd, closest_distances_in_hd = \
        gpu.calculate_knn_distances(samples_matrix=data, perplexity=perplexity, verbose=True)

    io.save_data_for_barneshut(files_dir, closest_distances_in_hd, closest_indices_in_hd, num_of_dims=num_dims,
                               perplexity=perplexity, theta=theta, eta=eta, exageration=exageration,
                               iterations=iterations, random_seed=random_seed, verbose=verbose)

    del samples

    exe_file = get_barnes_hut_executable(exe_dir)

    run_executable(exe_file, files_dir, verbose)

    tsne = io.load_tsne_result(files_dir)

    return tsne


def t_sne_from_existing_distances(files_dir, exe_dir=None, num_dims=2, perplexity=100, theta=0.4, eta=200,
                                  exageration=12.0, iterations=1000, random_seed=-1, verbose=2):
    """
    Runs the t-SNE algorithm using a precalculated set of distances (saved in the data.dat file in the files_dir folder)
    This will extract the distances and indices from the data.dat file and will overwrite it with the new parameters
    like num_dims, perplexity, etc..

    :param files_dir: where the data.dat file is and where the t-SNE will save its result
    :type files_dir: string
    :param exe_dir: where the Barnes_Hut.exe is (if None then the algorithm will try and find it itself)
    :type exe_dir: string
    :param num_dims: how many dimensions should the t-sne embedding have (2 or 3)
    :type num_dims: int
    :param perplexity: the number of closest spikes considered are 3 * perplexity
    :type perplexity: int
    :param theta: the angle on the t-sne embedding inside of which all spikes are considered a single point (0 to 1)
    :type theta: float
    :param eta: the learning rate of the algorithm
    :type eta: float
    :param exageration: the initial blow out (push out) of the points on the embedding
    :type exageration: float
    :param iterations: the number of iterations of the t-sne algorithm
    :type iterations: int
    :param random_seed: the random seed to create the initial random positions of the points on the embedding
    :type random_seed: float
    :param verbose: levels of verbosity. 0 is no output. If 3 then the algorithm will also save all intermediate t-sne
    embeddings. Useful for making videos of the t-sne process

    :type verbose: int
    :return: the t-NSE result
    :rtype: float32[:,num_of_dims]
    """
    closest_distances_in_hd, closest_indices_in_hd, parameters_dict = io.load_barneshut_data(files_dir,
                                                                                             data_has_exageration=False)

    io.save_data_for_barneshut(files_dir, closest_distances_in_hd, closest_indices_in_hd, num_of_dims=num_dims,
                               perplexity=perplexity, theta=theta, eta=eta, exageration=exageration,
                               iterations=iterations, random_seed=random_seed, verbose=verbose)

    exe_file = get_barnes_hut_executable(exe_dir)

    run_executable(exe_file, files_dir, verbose)

    tsne = io.load_tsne_result(files_dir)

    return tsne




import numpy as np
import os
import matplotlib.pyplot as plt


def _get_relevant_channels_with_threshold(threshold, template):
    amplitude = np.nanmax(template) - np.nanmin(template)
    points_over_threshold = np.argwhere(template > (np.nanmax(template) - threshold * amplitude))
    channels_over_threshold = np.unique(points_over_threshold[:, 1])
    return channels_over_threshold


def _get_relevant_channels_over_median_peaks(threshold, template):
    median = np.median(np.nanmin(template, axis=0))
    std = np.std(np.nanmin(template, axis=0))
    points_under_median = np.argwhere(template < (median - threshold*std))
    channels_over_threshold = np.unique(points_under_median[:, 1])
    return channels_over_threshold


def _normalize(L, normalizeFrom=0, normalizeTo=1):
    '''normalize values of a list to make its min = normalizeFrom and its max = normalizeTo'''
    vMax = max(L)
    vMin = min(L)
    return [(x-vMin)*(normalizeTo - normalizeFrom) / (vMax - vMin) for x in L]
# ------------------


def generate_probe_positions_of_spikes(base_folder, binary_data_filename, number_of_channels_in_binary_file,
                                       used_spikes_indices=None, position_mult=2.25, threshold=0.1):
    """
    Generate positions (x, y coordinates) for each spike on the probe. This function assumes that the spikes were
    generated with the kilosort algorithm so the base_folder holds all the necessary .npy arrays.
    In order for this function to find which channels are the most relevant in each spike it looks into the spike's
    assigned template (a channels x time points array in spike_templates.npy). It then find the minimum points of all
    channels, takes their median and their standard deviation and for each channel creates the difference between the
    minimum and the median. Finally it demarcates the relevant to the template channels by keeping the ones whose
    difference is larger than a number of times (threshold) over the standard deviation.
    It then picks the relevant channels of the spike's raw data, finds the differences between the minimum value
    and the channel's time series median value (over time), orders the channels according to these differences and
    assigns weights between 0 and 1 (0 for a difference of 0, 1 for a maximum difference).
    It finally finds the x, y positions of the selected channels and adds to the position of the largest difference
    channel the weighted average positions of the remaining selected channels

    :param base_folder: the folder name into which the kilosort result .npy arrays are
    :type base_folder: string
    :param binary_data_filename: the name of the binary file that holds the raw data that were originally passed to kilosort
    :type binary_data_filename: string
    :param number_of_channels_in_binary_file: How many channels does the binary file have (this is different to the number
    of channels that are set to active in kilosort)
    :type number_of_channels_in_binary_file: int
    :param used_spikes_indices: which of the spikes found by kilosort should be considered.
    :type used_spikes_indices: int[:]
    :param threshold: the number of times the standard deviation should be larger than the difference between a
    channel's minimum and the median of the minima of all channels in order to demarcate the channel as
    relevant to the spike

    :type threshold: float
    :return:  The position of each spike on the probe, the distance of eac h spike on the probe from the 0, 0 of the
    probe, the indices of the original ordering of the spikes on the new order sorted according to their distance on
    the probe, the distance of each spike on the probe sorted
     :rtype: float32[len(used_spike_indices), 2], float32[len(used_spike_indices)], int[len(used_spike_indices)], float32[len(used_spike_indices)]

    """
    # Load the required data from the kilosort folder
    channel_map = np.load(os.path.join(base_folder, 'channel_map.npy'))
    active_channel_map = np.squeeze(channel_map, axis=1)
    channel_positions = np.load(os.path.join(base_folder, 'channel_positions.npy'))

    spike_templates = np.load(os.path.join(base_folder, r'spike_templates.npy'))
    templates = np.load(os.path.join(base_folder, r'templates.npy'))

    data_raw = np.memmap(os.path.join(base_folder, binary_data_filename),
                         dtype=np.int16, mode='r')

    number_of_timepoints_in_raw = int(data_raw.shape[0] / number_of_channels_in_binary_file)
    data_raw_kilosorted = np.reshape(data_raw, (number_of_channels_in_binary_file, number_of_timepoints_in_raw), order='F')

    spike_times = np.squeeze(np.load(os.path.join(base_folder, 'spike_times.npy')).astype(np.int))

    time_points = 50
    if used_spikes_indices is None:
        used_spikes_indices = np.arange(0, len(spike_times))

    # Run the loop over all spikes to get the positions
    counter = 0
    weighted_average_postions = np.empty((len(used_spikes_indices), 2))
    spike_distance_on_probe = np.empty(len(used_spikes_indices))
    for spike_index in np.arange(0, len(used_spikes_indices)):
        spike_raw_data = data_raw_kilosorted[active_channel_map,
                                             (spike_times[used_spikes_indices[spike_index]]-time_points):
                                             (spike_times[used_spikes_indices[spike_index]]+time_points)]
        template = templates[spike_templates[used_spikes_indices[spike_index]], :, :].squeeze()
        relevant_channels = _get_relevant_channels_over_median_peaks(threshold, template)

        spike_raw_data_median_over_time = np.median(spike_raw_data, axis=1)
        peaks_to_median = spike_raw_data_median_over_time - spike_raw_data.min(axis=1)
        peaks_to_median = peaks_to_median[relevant_channels]

        relevant_channels_sorted = [v for (k, v) in sorted(zip(peaks_to_median, relevant_channels), reverse=True)]

        peaks_to_median_sorted = sorted(peaks_to_median, reverse=True)
        peaks_to_median_sorted.append(np.median(spike_raw_data_median_over_time[relevant_channels]))

        weights = _normalize(peaks_to_median_sorted)[:-1]
        relevant_channels_positions = channel_positions[relevant_channels_sorted]

        pos_x = relevant_channels_positions[0, 0]
        pos_y = relevant_channels_positions[0, 1]

        new_pos_x = pos_x - np.mean(((pos_x - relevant_channels_positions[:, 0]) * weights)[1:])
        new_pos_y = pos_y - np.mean(((pos_y - relevant_channels_positions[:, 1]) * weights)[1:])
        weighted_average_postions[spike_index, :] = [new_pos_x, new_pos_y]
        spike_distance_on_probe[spike_index] = np.sqrt(np.power(new_pos_x, 2) + np.power(new_pos_y, 2))

        counter += 1
        if counter % 5000 == 0:
            print('Completed ' + str(counter) + ' spikes')
    weighted_average_postions = weighted_average_postions * position_mult

    # sort according to position on probe
    spike_indices_sorted_by_probe_distance = np.array([b[0] for b in sorted(enumerate(spike_distance_on_probe),
                                                                            key=lambda dist: dist[1])])
    spike_distances_on_probe_sorted = np.array([b[1] for b in sorted(enumerate(spike_distance_on_probe),
                                                                     key=lambda dist: dist[1])])

    return weighted_average_postions, spike_distance_on_probe, \
        spike_indices_sorted_by_probe_distance, spike_distances_on_probe_sorted


def generate_probe_positions_of_templates(base_folder, threshold=0.1):
    """
    Generate positions (x, y coordinates) for each template found by kilosort on the probe.
    This function assumes that the base_folder holds all the necessary .npy arrays.
    In order for this function to find which channels are the most relevant in each template it looks into the
    template (a channels x time points array in spike_templates.npy). It then find the minimum points of all
    channels, takes their median and their standard deviation and for each channel creates the difference between the
    minimum and the median. Finally it demarcates the relevant to the template channels by keeping the ones whose
    difference is larger than a number of times (threshold) over the standard deviation.
    It then picks the relevant channels of the spike's raw data, finds the differences between the minimum value
    and the channel's time series median value (over time), orders the channels according to these differences and
    assigns weights between 0 and 1 (0 for a difference of 0, 1 for a maximum difference).
    It finally finds the x, y positions of the selected channels and adds to the position of the largest difference
    channel the weighted average positions of the remaining selected channels

    :param base_folder: the folder name into which the kilosort result .npy arrays are
    :type base_folder: string
    :param threshold: the number of times the standard deviation should be larger than the difference between a
    channel's minimum and the median of the minima of all channels in order to demarcate the channel asvrelevant to the
    spike

    :type threshold: float
    :return: weighted_average_postions : the positions of the templates on the probe
    :rtype: weighted_average_postions : float32[len(used_spike_indices) x 2]

    """
    # Load the required data from the kilosort folder
    channel_positions = np.load(os.path.join(base_folder, 'channel_positions.npy'))
    templates = np.load(os.path.join(base_folder, r'templates.npy'))
    template_markings = np.load(os.path.join(base_folder, r'template_marking.npy'))
    templates = templates[template_markings > 0, :, :]

    # Run the loop over all templates to get the positions
    counter = 0
    templates_positions = []
    for template in templates:
        relevant_channels = _get_relevant_channels_over_median_peaks(threshold, template)

        template_median_over_time = np.median(template, axis=0)
        peaks_to_median = template_median_over_time - template.min(axis=0)
        peaks_to_median = peaks_to_median[relevant_channels]

        relevant_channels_sorted = [v for (k, v) in sorted(zip(peaks_to_median, relevant_channels), reverse=True)]

        peaks_to_median_sorted = sorted(peaks_to_median, reverse=True)
        peaks_to_median_sorted.append(np.median(template_median_over_time[relevant_channels]))

        weights = _normalize(peaks_to_median_sorted)[:-1]
        relevant_channels_positions = channel_positions[relevant_channels_sorted]

        pos_x = relevant_channels_positions[0, 0]
        pos_y = relevant_channels_positions[0, 1]

        new_pos_x = pos_x - np.mean(((pos_x - relevant_channels_positions[:, 0]) * weights)[1:])
        new_pos_y = pos_y - np.mean(((pos_y - relevant_channels_positions[:, 1]) * weights)[1:])
        templates_positions.append([new_pos_x, new_pos_y])
        counter += 1
        if not (counter % 100):
            print('Completed ' + str(counter) + ' templates')

    return np.array(templates_positions)


def view_spike_positions(spike_positions, brain_regions, probe_dimensions, labels_offset=80, font_size=20):
    """
    Plot the spike positions as a scatter plot on a probe marked with brain regions

    :param spike_positions: the x,y positions of the spikes
    :type spike_positions: (np.array((N,2)))
    :param brain_regions: a dictionary with keys the names of the brain regions underneath the demarcating lines and
    values the y position on the probe of the demarcating lines

    :type brain_regions: dict
    :param probe_dimensions: the x and y limits of the probe
    :type probe_dimensions: (np.array(2))

    """
    fig = plt.figure()
    ax = fig.add_axes([0.08, 0.05, 0.9, 0.9])
    ax.scatter(spike_positions[:, 0], spike_positions[:, 1], s=5)
    ax.set_xlim(0, probe_dimensions[0])
    ax.set_ylim(0, probe_dimensions[1])
    ax.yaxis.set_ticks(np.arange(0, probe_dimensions[1], 100))
    ax.tick_params(axis='y', direction='in', length=5, width=1, colors='b')
    for region in brain_regions:
        ax.text(2, brain_regions[region] - labels_offset, region, fontsize=font_size)
        ax.plot([0, probe_dimensions[0]], [brain_regions[region], brain_regions[region]], 'k--', linewidth=2)
    return fig, ax





from os import path
from struct import calcsize, pack, unpack
from platform import system
import numpy as np


def _read_unpack(fmt, fh):
    return unpack(fmt, fh.read(calcsize(fmt)))


def find_exe_file():
    """
    Tries to find the Barnes_Hut executable given where it should be residing according to the OS and if this is installed
    as a package

    :return: the path to the executable
    :rtype: string
    """
    exe_dir = 'bin'
    exe_file = 'Barnes_Hut.out'
    if system() == 'Windows':
        exe_dir = 'Scripts'
        exe_file = 'Barnes_Hut.exe'

    dir_to_exe = None
    current_dir = path.dirname(__file__)
    while dir_to_exe == None:
        if path.isdir(path.join(current_dir, exe_dir)):
            dir_to_exe = path.join(current_dir, exe_dir)
        current_dir = path.dirname(current_dir)
        if len(current_dir) == 3:
            return None
    tsne_path = path.join(dir_to_exe, exe_file)

    return tsne_path


def save_data_for_barneshut(files_dir, sorted_distances, sorted_indices, num_of_dims, perplexity, theta, eta,
                            exageration, iterations, random_seed, verbose):
    """
    Save the pre calculated spike distances and the corresponding spike indices to disk in a file with a header carrying
    all the required paramteres for the barnes hut algorithm to run. The file's name will be data.dat
    This will be read by the barnes hut executable which will run the t-sne loop on the distances and produce the t-sne
    results.

    :param files_dir: the folder where the data.dat file will be saved
    :type files_dir: string
    :param sorted_distances: the distances of the nearest spikes to each spike (sorted)
    :type sorted_distances: float32[:, 3*perplexity +1]
    :param sorted_indices: the indicies of the nearest spikes to each spike (sorted)
    :type sorted_indices: float32[:, 3*perplexity +1]
    :param num_of_dims: how many dimensions should the t-sne embedding have (2 or 3)
    :type num_of_dims: int
    :param perplexity: the number of closest spikes considered are 3 * perplexity
    :type perplexity: int
    :param theta: the angle on the t-sne embedding inside of which all spikes are considered a single point (0 to 1)
    :type theta: float
    :param eta: the learning rate of the algorithm
    :type eta: float
    :param exageration: the initial blow out (push out) of the points on the embedding
    :type exageration: float
    :param iterations: the number of iterations of the t-sne algorithm
    :type iterations: int
    :param random_seed: the random seed to create the initial random positions of the points on the embedding
    :type random_seed: float
    :param verbose: levels of verbosity. 0 is no output. If 3 then the algorithm will also save all intermediate t-sne
    embeddings. Useful for making videos of the t-sne process

    :type verbose: int
    """
    sorted_indices = np.array(sorted_indices, dtype=np.int32)
    num_of_spikes = len(sorted_distances)
    num_of_nns = len(sorted_distances[0])

    filename = 'data.dat'

    with open(path.join(files_dir, filename), 'wb') as data_file:
        # Write the t_sne_bhcuda header
        data_file.write(pack('dddiiiiiii', theta, eta, exageration, num_of_spikes, num_of_dims, num_of_nns, iterations,
                             random_seed, verbose, perplexity))
        # Write the data
        for sample in sorted_distances:
            data_file.write(pack('{}d'.format(len(sample)), *sample))

        for sample in sorted_indices:
            data_file.write(pack('{}i'.format(len(sample)), *sample))


def load_tsne_result(files_dir, filename='result.dat'):
    """
    Load the results of the t-sne executable into a numpy array

    :param files_dir: the folder where the t-sne result file is
    :type files_dir: string
    :param filename: the filename of the results file
    :type filename: string
    :return: the t-sne results
    :rtype: float32[:, :]
    """
    # Read and pass on the results
    with open(path.join(files_dir, filename), 'rb') as output_file:
        # The first two integers are the number of samples and the dimensionality
        result_samples, result_dims = _read_unpack('ii', output_file)
        # Collect the results, but they may be out of order
        results = [_read_unpack('{}d'.format(result_dims), output_file) for _ in range(result_samples)]

        return np.array(results)


def load_barneshut_data(files_dir, filename='data.dat', data_has_exageration=True):
    """
    Load the spike distances and spike indices into numpy array from the data.dat file previously generated. Using this
    one can get the spike distances and indices and then resave them to a new data.dat file with a different parameters
    header to run t-sne again with different parameters.

    :param files_dir: the folder where the data file is
    :type files_dir: string
    :param filename: the name of the data file
    :type filename: string
    :param data_has_exageration: for backwards compatibility. old data files did not have the exageration parameter
    :type data_has_exageration: bool
    :return: the sorted distances of the closest spikes to each spike, the sorted indices of the closest spikes to each
    spike, the dictionary of the parameters used in this data file

    :rtype: float32[:,:], float32[:,:], dict
    """
    data_file = path.join(files_dir, filename)

    print('Loading previously calculated high dimensional distances')

    with open(data_file, 'rb') as output_file:
        if data_has_exageration:
            theta, eta, exageration, num_of_spikes, num_of_dims, num_of_nns, iterations, \
            random_seed, verbose, perplexity = _read_unpack('dddiiiiiii', output_file)

            parameters_dict = {'theta': theta, 'eta': eta, 'exageration': exageration,  'num_of_spikes': num_of_spikes,
                               'num_of_dims': num_of_dims, 'num_of_nns': num_of_nns, 'iterations': iterations,
                               'random_seed': random_seed, 'verbose': verbose, 'perplexity': perplexity}

        else:
            theta, eta, num_of_spikes, num_of_dims, num_of_nns, iterations, \
            random_seed, verbose, perplexity = _read_unpack('ddiiiiiii', output_file)

            parameters_dict = {'theta': theta, 'eta': eta, 'num_of_spikes': num_of_spikes,
                               'num_of_dims': num_of_dims, 'num_of_nns': num_of_nns, 'iterations': iterations,
                               'random_seed': random_seed, 'verbose': verbose, 'perplexity': perplexity}

        sorted_distances = np.array(
            [_read_unpack('{}d'.format(num_of_nns), output_file) for _ in range(num_of_spikes)])

        sorted_indices = np.array(
            [_read_unpack('{}i'.format(num_of_nns), output_file) for _ in range(num_of_spikes)])

    print('     Size of distances matrix: ' + str(sorted_distances.shape))
    print('     Size of indices matrix: ' + str(sorted_indices.shape))

    return sorted_distances, sorted_indices, parameters_dict





import numpy as np
from numba import cuda
import pyculib.blas as cublas
import pyculib.sorting as sorting
import numba
from numba import float32, guvectorize
import math
from . import stopwatch as sw

THREADS_PER_BLOCK = 32


@guvectorize([(float32[:, :], float32[:])], '(m,k)->(m)', nopython=True)
def _create_dot_product(a, dots_a):
    """
    Creates the dot product a[i] . a[i] for each row of the matrix and returns the vector of products

    :param a: The matrix
    :type a: float32[:,:]
    :param dots_a: The resulting vector of dot product squares
    :type dots_a: float32[:]
    """
    for i in np.arange(a.shape[0]):
        dots_a[i] = np.dot(a[i, :], a[i, :])


@cuda.jit
def _sums_of_dots_gpu(dots_a, dots_b, s_o_dots):
    """
    GPU kernel function creating the sums of all values of dots_a vector with all values of dots_b vector

    :param dots_a: The first vector
    :type dots_a: device float32[:]
    :param dots_b: The second vector
    :type dots_b: device float32[:]
    :param s_o_dots: The resulting sums
    :type s_o_dots: device float32[:,:]
    """
    a_index = cuda.threadIdx.x + cuda.blockDim.x * cuda.blockIdx.x

    if a_index < dots_a.shape[0]:
        for b_index in range(dots_b.shape[0]):
            s_o_dots[a_index, b_index] = dots_a[a_index] + dots_b[b_index]


def _get_inner_products(a, b, verbose):
    """
    Runs the _create_dot_product function on two matrices a and b and times itself

    :param a: The first matrix
    :type a: float32[:, :]
    :param b: The second matrix
    :type b: flaot32[:, :]
    :param verbose: If true print the time the function took to complete
    :type verbose: bool
    :return: The resulting vectors
    :rtype: (float32[:], float32[:])
    """
    with sw.Stopwatch() as stopwatch:
        dots_a = _create_dot_product(a)
        dots_b = _create_dot_product(b)

    if verbose:
        print("Making the dot products Time: {0:.3f} s".format(stopwatch.total_run_time))

    return dots_a, dots_b


def _get_sum_of_dot_products(dots_a, dots_b, distances_on_gpu, verbose):
    """
    Put the dot products on the gpu, calculate the ||a||^2 + ||b||^2 sum of dot products matrix (a.a + b.b) (on the gpu)
    and time the whole function

    :param dots_a: The first vector
    :type dots_a: float32[:]
    :param dots_b: The second vector
    :type dots_b: float32[:]
    :param distances_on_gpu: The resulting per element sums matrix
    :type distances_on_gpu: device float32[:, :]
    :param verbose: If true print the time required to complete the function
    :type verbose: bool
    """
    with sw.Stopwatch() as stopwatch:
        ddots_a = cuda.to_device(np.asfortranarray(dots_a))
        ddots_b = cuda.to_device(np.asfortranarray(dots_b))

        threads_per_block = THREADS_PER_BLOCK
        blocks_per_grid = math.ceil(ddots_a.shape[0] / threads_per_block)
        _sums_of_dots_gpu[blocks_per_grid, threads_per_block](ddots_a, ddots_b, distances_on_gpu)
        numba.cuda.synchronize()

    if verbose:
        print("Summing the dot products on the GPU Time: {0:.3f} s".format(stopwatch.total_run_time))


def _get_gpu_general_matrix_matrix_product(m, n, k, a, b, distances_on_gpu, verbose):
    """
    Calculate the -2<a,b> cross dot products matrix and do the sum (||a||^2 + ||b||^2) -2<a,b>

    :param m: First dimension of first matrix (number of samples in first matrix)
    :type m: int
    :param n: First dimension of second matrix (number of samples in second matrix)
    :type n: int
    :param k: Second dimension of both matrices (number of element in each sample)
    :type k: int
    :param a: First matrix
    :type a: float32[m, k]
    :param b: Second matrix
    :type b: float32[n, k]
    :param distances_on_gpu: Resulting distance matrix
    :type distances_on_gpu: float32[m, n]
    :param verbose: If true print the time the function took to complete
    :type verbose: bool
    """
    with sw.Stopwatch() as stopwatch:
        da = cuda.to_device(np.asfortranarray(a))
        db = cuda.to_device(np.asfortranarray(b))
        blas = cublas.Blas()
        blas.gemm('N', 'T', m, n, k, -2.0, da, db, 1.0, distances_on_gpu)
        numba.cuda.synchronize()

    if verbose:
        print("cuBLAS gemm Time: {0:.3f} s".format(stopwatch.total_run_time))


def _calculate_distances_on_gpu(a, b, distances_on_gpu, verbose=False):
    """
    Calculate the pairwise distances between all row vectors of matrix a and all row vectors of matrix b (on the gpu)

    :param a: First matrix
    :type a: float32[m, k]
    :param b: Second matrix
    :type b: float32[n, k]
    :param distances_on_gpu: The resulting distances between m vectors of matrix one and n vectors of matrix two
    :type distances_on_gpu: device float32[m, n]
    :param verbose: If true print out the times of all steps in this function
    :type verbose: bool
    """
    m = a.shape[0]
    n = b.shape[0]
    k = a.shape[1]

    dots_a, dots_b = _get_inner_products(a, b, verbose)

    _get_sum_of_dot_products(dots_a, dots_b, distances_on_gpu, verbose)

    _get_gpu_general_matrix_matrix_product(m, n, k, a, b, distances_on_gpu, verbose)


def _segment_sort_transposed_distances_get_knns(num_of_neighbours, distances_on_gpu, number_of_sorts,
                                                verbose=False):
    """
    Sorts the distances_on_gpu matrix (on the gpu) and for each element returns the num_of_neighbours smallest
    distances with their position on the vector of elements (indices)

    :param num_of_neighbours: Number of nearest neighbours to each element to return
    :type num_of_neighbours: int
    :param distances_on_gpu: The full, unsorted, pairwise elements distances
    :type distances_on_gpu: device float32[m , m]
    :param number_of_sorts: Given the GPU's memory, in how many pieces must the algorithm break the given distance matrix and sort each piece
    :type number_of_sorts: int
    :param verbose: If true print some info
    :type verbose: bool
    :return: The num_of_neighbors nearest elements' indices and distances vectors
    :rtype: float32[m, k], float32[m, k]
    """
    m = distances_on_gpu.shape[0]  # all spikes
    n = distances_on_gpu.shape[1]  # part of spikes in this distances_on_gpu matrix

    selected_sorted_distances = np.empty((n, num_of_neighbours))
    selected_sorted_indices = np.empty((n, num_of_neighbours))

    with sw.Stopwatch() as stopwatch:
        # Create an array with the indices each iteration will step through
        p = np.append(np.arange(0, n, int(n / number_of_sorts)), n)
        for i in np.arange(1, p.shape[0]):
            # Get how many elemnts each iteration will do
            delta_n = p[i] - p[i - 1]

            # Prepare for sorting in this iteration
            # 1. Get this iteration's keys (distances) onto the host from the gpu and reshape them appropriately
            keys = np.ascontiguousarray(distances_on_gpu.copy_to_host()[:, p[i - 1]:p[i]].transpose().reshape(m * delta_n))
            # 2. Create the values (indices) appropriately reshaped
            values = np.ascontiguousarray(np.tile(np.arange(m), (delta_n, 1)).reshape(m * delta_n))
            # 3. Define the segments
            segments = np.ascontiguousarray(np.arange(m, m * delta_n, m))
            # 4. Do the segmented sort
            sorting.segmented_sort(keys=keys, vals=values, segments=segments)

            # Put the smallest num_of_neighbours distances (and indices) in the selected_sorted_distances
            # (selected_sorted_indices)
            keys = np.reshape(keys, (delta_n, m))[:, :num_of_neighbours]
            values = np.reshape(values, (delta_n, m))[:, :num_of_neighbours]
            selected_sorted_distances[p[i - 1]:p[i], :] = keys[:, :]
            selected_sorted_indices[p[i - 1]:p[i], :] = values[:, :]
            if verbose:
                print('     Sorted {} of {} segments of this iteration'.format(i, p.shape[0] - 1))

    print("SORTING TIME: {0:.3f} s".format(stopwatch.total_run_time))

    return selected_sorted_indices, selected_sorted_distances


def _get_required_gpu_memory():
    """
    Return the number of bytes available in the GPU to do all following steps
    :return: The number of GPU bytes
    :rtype: int
    """
    gpu_mem = cuda.current_context().get_memory_info()
    available_gpu_mem = 0.5 * gpu_mem[0]

    return available_gpu_mem


def calculate_knn_distances(samples_matrix, perplexity=100, verbose=True):
    """
    Calculates the k (perplexity * 3 + 1) nearest neighbors of all row vectors (samples) in the sample_matrix. It uses
    the GPU to do this so it has to take into account the available GPU memory. Given that, it does the comparison in
    steps where the first matrix is the full samples matrix and the second matrix is a sub set of it. Each step first
    calculates all distances between the row vectors of the first and second matrix and then sorts these distances and
    keep the k closest ones.

    :param samples_matrix: The full smaples x dimensions matrix where every sample is a row in the matrix
    :type samples_matrix: float32[m, n]
    :param perplexity: The definition of how many nearest neighbors to keep (k = perplexity * 3 + 1)
    :type perplexity: int
    :param verbose: If true the function prints out the time takes to complete the different parts of it
    :type verbose: bool
    :return: The num_of_neighbors nearest elements' indices and distances vectors
    :rtype: float32[m, k], float32[m, k]
    """
    with sw.Stopwatch() as outside_stopwatch:
        num_of_neighbours = perplexity * 3 + 1  # That is how the original t-sne defined k nearest neighbours
        m = samples_matrix.shape[0]  # number of samples in the samples x dimensions matrix

        closest_indices = np.empty((m, num_of_neighbours))
        closest_distances = np.empty((m, num_of_neighbours))

        available_gpu_mem = _get_required_gpu_memory()

        n = int(np.ceil(available_gpu_mem / (4 * m)))  # The number of samples to compare in each iteration with the full m
        # samples (defined by the GPU's available memory)

        number_of_iters = int(np.ceil(m / n))  # The number of iterations so that all m samples have their k nearest
        # neighbours with all other m - 1 samples calculated

        # Generate the pairs of star and end indices for the samples that will be included in the second matrix at every
        # iteration
        indices_of_second_matrices = [(i, i + n) for i in np.arange(0, number_of_iters * (n - 1), n)]
        indices_of_second_matrices[-1] = (indices_of_second_matrices[-1][0], m)

        first_matrix = np.array(samples_matrix, dtype=np.float32)  # first matrix is the full samples matrix

        # Start the loop over iterations filling up the closest_indices and closest_distances matrices
        for iteration in np.arange(number_of_iters):
            # Fill up the second matrix with the correct part of the whole samples matrix
            second_matrix = np.array(samples_matrix[indices_of_second_matrices[iteration][0]:
                                                    indices_of_second_matrices[iteration][1], :],
                                     dtype=np.float32)

            with sw.Stopwatch() as inside_stopwatch:
                if iteration != 0:
                    del distances_on_gpu
                cuda.current_context().deallocations.clear()

                if verbose:
                    print('LOADING UP THE GPU')
                temp = np.array(np.zeros((m, second_matrix.shape[0]), dtype=np.float32))
                distances_on_gpu = cuda.to_device(np.asfortranarray(temp))

                if verbose:
                    print("Loading matrix time:  {0:.3f} s".format(inside_stopwatch.time_elapsed))

                if verbose:
                    print('ITERATION NUMBER: ' + str(iteration + 1))

                _calculate_distances_on_gpu(a=first_matrix, b=second_matrix, distances_on_gpu=distances_on_gpu,
                                            verbose=verbose)

                available_gpu_mem = _get_required_gpu_memory()

                number_of_sorts = int(np.ceil((16 * n * m) / available_gpu_mem))  # 4 is the bytes per float32,
                # 2 is the two arrays that need to be loaded to gpu, the other factor of 2 is probably a doubling
                # overhead in the algorithm

                if verbose:
                    print('     Number of sorting segments = ' + str(number_of_sorts + 1))

                temp_indices, temp_distances = \
                    _segment_sort_transposed_distances_get_knns(num_of_neighbours=num_of_neighbours,
                                                                distances_on_gpu=distances_on_gpu,
                                                                number_of_sorts=number_of_sorts, verbose=verbose)

                closest_indices[indices_of_second_matrices[iteration][0]: indices_of_second_matrices[iteration][1], :] = \
                    np.ascontiguousarray(temp_indices)
                closest_distances[indices_of_second_matrices[iteration][0]: indices_of_second_matrices[iteration][1], :] = \
                    np.ascontiguousarray(temp_distances)
                if verbose:
                    print('FINISHED CALCULATING ' + str(iteration + 1) + ' OF ' + str(number_of_iters) +
                          ' ITERATIONS')

            if verbose:
                print("Spend Time: {0:.3f} s".format(outside_stopwatch.time_elapsed))

    return closest_indices, np.sqrt(np.abs(closest_distances))




"""
Set of functions that allow the t-sne algorithm to use the result of the kilosort automatic spike clustering algorithm.
The functions assume file names in the base_folder (the folder that kilosort puts its results in) as generated by
kilosort.

Some functions from this collection are also useful together with the GUI that allows cleaning of the kilosort templates
and the GUI that allows manual curation of the kilosort results using the t-sne embedding (e.g. generate_spike_info)
"""


import numpy as np
from os import path
import pickle
import pandas as pd
from . import io_with_cpp as io


def spike_indices_of_template(spike_templates, clean_templates=None, clean_template_index=None):
    """
    Returns the indices of the spikes belonging to a template from the group of clean templates (after manual cleaning
    of the kilosort results). The manual cleaning results are pulled from the template_markings.npy file saved the
    cleaning GUI in the same folder as the kilosort results

    :param spike_templates: The vector in the spike_templates.npy file of the kilosort results
    :type spike_templates: int[:]
    :param clean_templates: The vector of indices in the vector of the original templates of the templates that are
    clean. E.g. If originally templates [1,2,4,5,13,22,23] where found but templates 2 and 13 where noise then
    clean_templates = [0,2,3,4,6,7]

    :type clean_templates: int[:]
    :param clean_template_index: The index in the vector of the clean_templates of the template whose spikes we need
    :type clean_template_index: int
    :return: a vector of the indices of spikes belonging to the template
    :rtype: int[:]
    """
    if clean_templates is None:
        original_template_index = spike_templates
    else:
        original_template_index = clean_templates[clean_template_index]
    spikes_indices = np.squeeze(np.argwhere(np.in1d(spike_templates, original_template_index)))
    return spikes_indices


def find_templates_with_number_of_spikes_under_threshold(spike_templates, clean_templates, threshold):
    """
    Finds all the templates that have fewer number of spikes than a given threshold

    :param spike_templates: the original spike_templates as generated by kilosort
    :type spike_templates: int[:]
    :param clean_templates: the vector of which template each spike is assigned to but only for the clean templates
    :type clean_templates: int[:]
    :param threshold: the number of spikes allowed in a chosen template
    :type threshold: int
    :return: a dictionary with keys the templates with fewer than threshold spikes (small templates) and values their
    spike indices, the total number of spikes in these small templates, a dictionary with keys the remaining (large)
    templates and values the indices of their spikes

    :rtype: dict, int , dict
    """
    small_clean_templates_with_indices = dict()
    large_clean_templates_with_indices = dict()
    num_of_spikes_in_small_templates = 0
    for i in np.arange(len(clean_templates)):
        spike_indices = spike_indices_of_template(spike_templates, clean_templates, i)
        num_of_spikes = spike_indices.size
        if num_of_spikes <= threshold:
            small_clean_templates_with_indices[i] = spike_indices
            num_of_spikes_in_small_templates += num_of_spikes
        else:
            large_clean_templates_with_indices[i] = spike_indices

    return small_clean_templates_with_indices, num_of_spikes_in_small_templates, large_clean_templates_with_indices


def get_template_marking(base_folder):
    """
    Returns the template_marking vector (if a manual curation session has been performed on the kilosort results) which
    assigns to each template one of 7 possibilities. Since 0 means noise and all other possibilities are positive
    numbers this vector can act as a mask to exclude spikes of templates that have been marked as noise.
    If no template_markings.npy file is found then the function returns a vector of ones as long as the spike_templates
    vector

    :param base_folder: the folder with the kilosort results
    :type base_folder: string
    :return: vector of markings for each spike according to the type its template has been manually assigned to
    :rtype: int[:]
    """

    if path.isfile(path.join(base_folder, 'template_marking.npy')):
        template_marking = np.load(path.join(base_folder, 'template_marking.npy'))
    else:
        templates = np.load(path.join(base_folder, 'templates.npy'))
        num_of_templates = templates.shape[2]
        template_marking = np.ones(num_of_templates)
        del templates
        print('No template_marking.npy found. Using all templates.')

    return template_marking


def find_spike_indices_for_representative_tsne(base_folder, save_to_folder, threshold, total_spikes_required):
    """
    The function is used to create a subset of spikes using the following criteria. 1) The total number of spikes in the
    subset has to be smaller than the total_spikes_required value. 2) All spikes belonging to templates with threshold
    or fewer number of spikes should be included. If the threshold is too big and the total_spikes_required too small
    a number then these might not be mutually achievable criteria. If this is not the case then all the spikes belonging
    to templates with fewer than threshold spikes will be included. That will give a total of spikes smaller than the
    total_spikes_required value. The difference between the total_spikes_required and the total spikes in the small
    templates will be filled by spikes from the remaining large templates (the ones with higher number of spikes than
    threshold). The algorithm will calculate how many spikes these large templates have and use this to calculate a
    percentage of spikes for each large template that will bring the number of spikes in the subset equal to the
    total_spikes_required value. In this way the subset will be a full representation of the small templates (which
    will be the large majority of templates) and a random representation of part of the larger templates (which will be
    the small minority). Each large spike will contribute a specific percentage of its own total spikes.

    This technique can be used to create subsets of spikes whose total number of spikes does not exceed the t-sne
    capabilities. Use the indices_of_spikes_used returned vector to pass to the spikes_used_with_clean_indexing
    parameter of the calculate_template_features_matrix_for_tsne function.
    After running t-sne on this set any large templates that are still of interest (have not been
    delegated to noise and after any merging and splitting) can be run by themselves on a separate t-sne.

    :param base_fofder: the folder with the kilosort results
    :type base_folder: string
    :param save_to_folder: the function except returning the results will also save them as files to this folder. It
    also save dictionaries that make searching through the small and large templates easier

    :type save_to_folder: string
    :param threshold: the number of spikes that full included templated must be under
    :type threshold: int
    :param total_spikes_required: the total number of spikes in the resulting subset
    :type total_spikes_required: int
    :return: a vector with the indices of all the spikes in the subset,
    a dictionary with keys the small templates and values their spike indices,
    a dictionary with keys the large templates and values their spike indices

    :rtype: int[:], dict, dict
    """
    spike_templates = np.load(path.join(base_folder, 'spike_templates.npy'))
    template_marking = get_template_marking(base_folder)

    clean_templates = np.argwhere(template_marking)

    small_clean_templates_with_spike_indices, num_of_spikes_in_small_templates, large_clean_templates_with_spike_indices = \
        find_templates_with_number_of_spikes_under_threshold(spike_templates, clean_templates, threshold)

    extra_spikes_required = total_spikes_required - num_of_spikes_in_small_templates

    spikes_clean_index = np.squeeze(np.argwhere(np.in1d(spike_templates, clean_templates)))
    percentage_of_kept_spikes_in_large_templates = extra_spikes_required / (spikes_clean_index.size -
                                                                            num_of_spikes_in_small_templates)

    spikes_chosen = []
    num_of_spikes_in_large_templates = 0
    for large_template_index in large_clean_templates_with_spike_indices.keys():
        spike_indices = large_clean_templates_with_spike_indices[large_template_index]
        num_of_spikes = spike_indices.size
        num_of_spikes_in_large_templates += num_of_spikes
        chosen_num_of_spikes = int(num_of_spikes * percentage_of_kept_spikes_in_large_templates)
        chosen_spike_indices = np.random.choice(spike_indices, chosen_num_of_spikes, replace=False)
        spikes_chosen.append(chosen_spike_indices)

    num_of_spikes_in_small_templates = 0
    for small_template_index in small_clean_templates_with_spike_indices.keys():
        spike_indices = small_clean_templates_with_spike_indices[small_template_index]
        spikes_chosen.append(spike_indices)
        num_of_spikes_in_small_templates += spike_indices.size

    print('{} templates with more than {} spikes (total spikes in those = {}). \n{} templates with less '
          'than {} spikes (total spikes in those = {})'.format(len(large_clean_templates_with_spike_indices),
                                                               threshold, num_of_spikes_in_large_templates,
                                                               len(small_clean_templates_with_spike_indices),
                                                               threshold, num_of_spikes_in_small_templates))

    spikes_chosen_flat = []
    for sublist in spikes_chosen:
        if np.size(sublist) > 1:
            for item in sublist:
                spikes_chosen_flat.append(item)
        else:
            spikes_chosen_flat.append(sublist)

    indices_of_spikes_used = np.array(spikes_chosen_flat)
    small_clean_templates_indices = np.fromiter(small_clean_templates_with_spike_indices.keys(), int,
                                                len(small_clean_templates_with_spike_indices))
    large_clean_tempalates_indices = np.fromiter(large_clean_templates_with_spike_indices.keys(), int,
                                                 len(large_clean_templates_with_spike_indices))
    pickle.dump(small_clean_templates_with_spike_indices,
                open(path.join(save_to_folder, "small_clean_templates_with_spike_indices.pkl"), "wb"))
    pickle.dump(large_clean_templates_with_spike_indices,
                open(path.join(save_to_folder, "large_clean_templates_with_spike_indices.pkl"), "wb"))
    np.save(path.join(save_to_folder, 'indices_of_spikes_used'), indices_of_spikes_used)
    np.save(path.join(save_to_folder, 'indices_of_small_templates'), small_clean_templates_indices)
    np.save(path.join(save_to_folder, 'indices_of_large_templates'), large_clean_tempalates_indices)

    return indices_of_spikes_used, small_clean_templates_with_spike_indices, large_clean_templates_with_spike_indices


def calculate_template_features_matrix_for_tsne(base_folder, save_to_folder, spikes_used_with_original_indexing=None,
                                                spikes_used_with_clean_indexing=None):
    """
    Using the kilosort results, this function creates a matrix of samples x elements that can be used as an input to
    the t-sne algorithm. Each sample represents a spike and each element the distance of this spike to a specific
    template.

    :param base_folder:
    :type base_folder:
    :param save_to_folder:
    :type save_to_folder:
    :param spikes_used_with_original_indexing:
    :type spikes_used_with_original_indexing:
    :param spikes_used_with_clean_indexing:
    :type spikes_used_with_clean_indexing:
    :return:
    :rtype:
    """
    spike_templates = np.load(path.join(base_folder, 'spike_templates.npy'))
    template_features = np.load(path.join(base_folder, 'template_features.npy'))
    template_features_ind = np.load(path.join(base_folder, 'template_feature_ind.npy'))

    template_marking = get_template_marking(base_folder)

    clean_templates = np.argwhere(template_marking)
    spikes_clean_index = np.squeeze(np.argwhere(np.in1d(spike_templates, clean_templates)))

    if spikes_used_with_original_indexing is not None and spikes_used_with_clean_indexing is not None:
        print('Use one of the spikes_used_... variable')
        return None
    elif spikes_used_with_original_indexing is None and spikes_used_with_clean_indexing is None:
        pass
    elif spikes_used_with_original_indexing is None and spikes_used_with_clean_indexing is not None:
        spikes_clean_index = spikes_clean_index[spikes_used_with_clean_indexing]
    elif spikes_used_with_original_indexing is not None and spikes_used_with_clean_indexing is None:
        spikes_clean_index = spikes_used_with_original_indexing

    clean_templates = np.unique(spike_templates[spikes_clean_index])

    template_features_sparse_clean = np.zeros((spikes_clean_index.size, clean_templates.size))
    s = 0
    for spike in spikes_clean_index:
        cluster = spike_templates[spike][0]
        indices = template_features_ind[cluster, :]
        if s % 5000 == 0:
            print('Spikes completed: {}'.format(s))
        s+=1
        for i in np.arange(len(indices)):
            template_features_sparse_clean[np.argwhere(spikes_clean_index == spike),
                                           np.argwhere(clean_templates == indices[i])] = template_features[spike, i]

    np.save(path.join(save_to_folder, 'data_to_tsne_' + str(template_features_sparse_clean.shape) + '.npy'),
            template_features_sparse_clean)

    return template_features_sparse_clean


def load_template_features_matrix_for_tsne(save_to_folder, shape):
    """
    Loads the npy array that carries the sparse matrix of spikes x distances to templates

    :param save_to_folder: the folder the matrix has been saved to
    :type save_to_folder: string
    :param shape: the shape of the matrix
    :type shape: tuple
    :return: the sparse matrix
    :rtype: float32[shape]
    """
    template_features_sparse_clean = np.load(path.join(save_to_folder, r'data_to_tsne_' + str(shape) + '.npy'))
    return template_features_sparse_clean


def generate_spike_info(kilosort_folder, tsne_folder, tsne_filename='result.dat'):
    """
    spike_info.df is a pandas DataFrame that collates all the information about a group of spikes for sorting purposes.
    It is also the data frame that the GUI to manually curate kilosorted and t-sne data uses to load and save its
    results.

    :param kilosort_folder: the folder of the kilosort results
    :type kilosort_folder: string
    :param tsne_folder: the folder with the t-sne results
    :type tsne_folder: string
    :param tsne_filename: the filename of the t-sne results numpy array
    :type tsne_filename: string
    :return: saves and returns the spike_info dataframe
    :rtype: pandas.DataFrame
    """
    tsne = io.load_tsne_result(tsne_folder, tsne_filename)

    partial_tsne = False

    if path.isfile(path.join(tsne_folder, 'indices_of_spikes_used.npy')):
        spikes_used = np.load(path.join(tsne_folder, 'indices_of_spikes_used.npy'))
        partial_tsne = True
    else:
        spikes_used = np.arange(tsne.shape[0])

    if path.isfile(path.join(tsne_folder, 'indices_of_small_templates.npy')) and partial_tsne:
        indices_of_small_templates = np.load(path.join(tsne_folder, 'indices_of_small_templates.npy'))
    else:
        indices_of_small_templates = spikes_used

    if path.isfile(path.join(tsne_folder, 'weighted_template_positions.npy')):
        weighted_template_positions = np.load(path.join(tsne_folder, 'weighted_template_positions.npy'))
    else:
        weighted_template_positions = None

    if path.isfile(path.join(tsne_folder, 'weighted_spike_positions.npy')):
        weighted_spike_positions = np.load(path.join(tsne_folder, 'weighted_spike_positions.npy'))
    else:
        weighted_spike_positions = None

    if path.isfile(path.join(tsne_folder, 'template_marking.npy')):
        template_marking = np.load(path.join(kilosort_folder, 'template_marking.npy'))
    else:
        template_marking = np.ones(tsne.shape[0]) * 5 # Set it to Unspecified_1

    spike_templates = np.load(path.join(kilosort_folder, 'spike_templates.npy'))[spikes_used]
    spike_times = np.load(path.join(kilosort_folder, 'spike_times.npy'))[spikes_used]

    types = {0: 'Noise', 1: 'SS', 2: 'SS_Contaminated', 3: 'SS_Putative', 4: 'MUA', 5: 'Unspesified_1',
             6: 'Unspecified_2',
             7: 'Unspecified_3'}

    columns = ['original_index', 'times', 'template_after_cleaning', 'type_after_cleaning', 'template_after_sorting',
               'type_after_sorting', 'template_with_all_spikes_present', 'tsne_x', 'tsne_y', 'probe_position_x',
               'probe_position_z']

    spike_info = pd.DataFrame(index=np.arange(spikes_used.size), columns=columns)

    spike_info['original_index'] = spikes_used
    spike_info['times'] = spike_times
    spike_info['template_after_cleaning'] = spike_templates
    spike_info['type_after_cleaning'] = [types[int(template_marking[i])] for i in spike_templates]
    spike_info['template_after_sorting'] = spike_info['template_after_cleaning']
    spike_info['type_after_sorting'] = spike_info['type_after_cleaning']
    if partial_tsne:
        spike_info['template_with_all_spikes_present'] = [bool(np.in1d(spike_template, indices_of_small_templates))
                                                          for spike_template in spike_templates]
    else:
        spike_info['template_with_all_spikes_present'] = [True for spike_template in spike_templates]

    if weighted_spike_positions is not None:
        spike_info['probe_position_x'] = weighted_spike_positions[:, 0]
        spike_info['probe_position_y'] = weighted_spike_positions[:, 1]

    if weighted_spike_positions is None and weighted_template_positions is not None:
        spike_info['probe_position_x'] = [weighted_template_positions[spike_template, 0]
                                          for spike_template in spike_templates]
        spike_info['probe_position_y'] = [weighted_template_positions[spike_template, 1]
                                          for spike_template in spike_templates]

    spike_info['tsne_x'] = tsne[:, 0]
    spike_info['tsne_y'] = tsne[:, 1]

    spike_info.to_pickle(path.join(tsne_folder, 'spike_info.df'))

    return spike_info



from setuptools import setup

setup(name='spikesorting_tsne',
      version='1.0.0',
      description='T-sne with burnes hut and cuda extension (with python wrappers and python code for spike sorting)',
      long_description='This is the repository for the python part of the conda python package that allows running the results of automated spike sorting algorithms through the t-SNE algorithm obtaining a 2D or 3D embedding of the spikes. \
						Although this package offers some functionality dedicated to spikesorting the t-SNE part of it is kept separate and can be run with any matrix of samples x features.\
						The package is split into two parts. The python part (in this repo) has the following functionality:\
						1. It has the main function through which the whole algorithm is called.\
						2. The main function wraps the C++ executable (Barnes_Hut.exe) and calls it appropriately.\
						3. It uses numba gpu to create the spike (or any other sample matrix) distances.\
						4. It offers functions that operate on the result of the kilosort spikesorting algorithm to create the required samples (spikes) x features (template distances) matrix for t-SNE to run on.\
						5. It offers functions that allow the user to split a dataset of spikes that is too lrage for the algorithm to sub groups that can be seperately run and then recombined (uses information generated by ythe kilosort algorithm)\
						The C++ part (which generates the Barnes_Hut.exe executable) can be found [here](https://github.com/georgedimitriadis/spikesorting_tsne_bhpart).\
						More detailed documentation can be found in the [Github Pages](https://georgedimitriadis.github.io/spikesorting_tsne/) of this repo.',
      long_description_content_type='text/markdown',
      url='https://github.com/georgedimitriadis/spikesorting_tsne',
      author='George Dimitriadis',
      author_email='gdimitri@hotmail.com',
      packages=['spikesorting_tsne'],
      classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Topic :: Scientific/Engineering :: Visualization',
        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    project_urls={
        'Documentation': 'https://georgedimitriadis.github.io/spikesorting_tsne/',
        'Python code repository': 'https://github.com/georgedimitriadis/spikesorting_tsne',
        'C++ code repository': 'https://github.com/georgedimitriadis/spikesorting_tsne_bhpart',
    },
    python_requires='>=3.5',
    zip_safe=False)


